# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from numba import njit
from ..utils import Cache
from .equilibrium import Equilibrium
from ..exceptions import NoEquilibrium
from .binary_phase_fraction import phase_fraction
from scipy.optimize import shgo, differential_evolution
import flexsolve as flx
import numpy as np

__all__ = ('LLE', 'LLECache')

# TODO: SUBMIT ISSUE TO NUMBA
@njit(cache=True)
def liquid_activities(mol_L, T, f_gamma, gamma_args):
    total_mol_L = mol_L.sum()
    x = mol_L / total_mol_L
    gamma = f_gamma(x, T, *gamma_args)
    xgamma = x * gamma
    return xgamma

@njit(cache=True)
def gibbs_free_energy_of_liquid(mol_L, xgamma):
    xgamma[xgamma <= 0] = 1
    g_mix = (mol_L * np.log(xgamma)).sum()
    return g_mix

@njit(cache=True)
def lle_objective_function(mol_L, mol, T, f_gamma, gamma_args):
    mol_l = mol - mol_L
    if mol_l.sum() == 0.:
        g_mix_l = 0.
    else:
        xgamma_l = liquid_activities(mol_l, T, f_gamma, gamma_args)
        g_mix_l = gibbs_free_energy_of_liquid(mol_l, xgamma_l)
    if mol_L.sum() == 0.:
        g_mix_L = 0.
    else:
        xgamma_L = liquid_activities(mol_L, T, f_gamma, gamma_args)
        g_mix_L = gibbs_free_energy_of_liquid(mol_L, xgamma_L)
    g_mix = g_mix_l + g_mix_L
    return g_mix

@njit(cache=True)
def psuedo_equilibrium_inner_loop(logKgammay, z, T, n, f_gamma, gamma_args, phi):
    logKgammay_new = logKgammay.copy()
    K = np.exp(logKgammay[:n])
    x = z/(1. + phi * (K - 1.))
    x[x < 0] = 1e-16
    x = x / x.sum()
    gammay = logKgammay[n:]
    gammax = f_gamma(x, T, *gamma_args)
    K = gammax / gammay 
    y = K * x
    y /= y.sum()
    gammay = f_gamma(y, T, *gamma_args)
    K = gammax / gammay
    logKgammay_new[:n] = np.log(K)
    logKgammay_new[n:] = gammay
    return logKgammay_new

def pseudo_equilibrium_outer_loop(logKgammayphi, z, T, n, f_gamma, gamma_args, inner_loop_options):
    logKgammayphi_new = logKgammayphi.copy()
    logKgammay = logKgammayphi[:-1]
    phi = logKgammayphi[-1]
    args=(z, T, n, f_gamma, gamma_args)
    logKgammay = flx.aitken(
        psuedo_equilibrium_inner_loop, logKgammay, 
        args=(*args, phi), **inner_loop_options,
    )
    K = np.exp(logKgammay[:n])
    try:
        phi = phase_fraction(z, K, phi)
    except (ZeroDivisionError, FloatingPointError):
        raise NoEquilibrium
    if np.isnan(phi): raise NoEquilibrium
    if phi > 1: phi = 1 - 1e-16
    if phi < 0: phi = 1e-16
    logKgammayphi_new[:-1] = logKgammay
    logKgammayphi_new[-1] = phi
    return logKgammayphi_new

def pseudo_equilibrium(K, phi, z, T, n, f_gamma, gamma_args, inner_loop_options, outer_loop_options):
    phi = phase_fraction(z, K, phi)
    try:
        x = z/(1. + phi * (K - 1.))
    except:
        x = np.ones(n)
    x /= x.sum()
    y = K * x
    logKgammayphi = np.zeros(2*n + 1)
    logKgammayphi[:n] = np.log(K)
    logKgammayphi[n:-1] = f_gamma(y, T, *gamma_args)
    logKgammayphi[-1] = phi
    try:
        logKgammayphi = flx.aitken(
            pseudo_equilibrium_outer_loop, logKgammayphi,
            args=(z, T, n, f_gamma, gamma_args, inner_loop_options),
            **outer_loop_options,
        )
    except NoEquilibrium:
        return z
    phi = logKgammayphi[-1]
    if 0 < phi < 1:
        K = np.exp(logKgammayphi[:n])
        return z/(1. + phi * (K - 1.)) * (1 - phi)
    else:
        return z

class LLE(Equilibrium, phases='lL'):
    """
    Create a LLE object that performs liquid-liquid equilibrium when called.
    The SHGO (simplicial homology global optimization) alogorithm [1]_ is used to find the 
    solution that globally minimizes the gibb's free energy of both phases.
        
    Parameters
    ----------
    imol=None : :class:`~thermosteam.indexer.MaterialIndexer`, optional
        Molar chemical phase data is stored here.
    thermal_condition=None : :class:`~thermosteam.ThermalCondition`, optional
        The temperature and pressure used in calculations are stored here.
    thermo=None : :class:`~thermosteam.Thermo`, optional
        Themodynamic property package for equilibrium calculations.
        Defaults to `thermosteam.settings.get_thermo()`.
    
    Examples
    --------
    >>> from thermosteam import indexer, equilibrium, settings
    >>> settings.set_thermo(['Water', 'Ethanol', 'Octane', 'Hexane'], cache=True)
    >>> imol = indexer.MolarFlowIndexer(
    ...             l=[('Water', 304), ('Ethanol', 30)],
    ...             L=[('Octane', 40), ('Hexane', 1)]
    ... )
    >>> lle = equilibrium.LLE(imol)
    >>> lle(T=360, top_chemical='Octane')
    >>> lle
    LLE(imol=MolarFlowIndexer(
            L=[('Water', 2.671), ('Ethanol', 2.284), ('Octane', 39.92), ('Hexane', 0.9885)],
            l=[('Water', 301.3), ('Ethanol', 27.72), ('Octane', 0.07884), ('Hexane', 0.01154)]),
        thermal_condition=ThermalCondition(T=360.00, P=101325))
    
    References
    ----------
    .. [1] Endres, SC, Sandrock, C, Focke, WW (2018) “A simplicial homology 
           algorithm for lipschitz optimisation”, Journal of Global Optimization.
    
    """
    __slots__ = ('composition_cache_tolerance',
                 'temperature_cache_tolerance',
                 'method',
                 '_z_mol',
                 '_T',
                 '_lle_chemicals',
                 '_IDs',
                 '_K',
                 '_gamma_y',
                 '_phi'
    )
    default_method = 'pseudo equilibrium'
    shgo_options = dict(f_tol=1e-6, minimizer_kwargs=dict(f_tol=1e-6))
    differential_evolution_options = {'seed': 0,
                                      'popsize': 12,
                                      'tol': 1e-6}
    pseudo_equilibrium_outer_loop_options = dict(
        xtol=1e-12, maxiter=100, checkiter=False, 
        checkconvergence=False, convergenceiter=10,
    )
    pseudo_equilibrium_inner_loop_options = dict(
        xtol=1e-9, maxiter=20, checkiter=False,
        checkconvergence=False, convergenceiter=5,
    )
    default_composition_cache_tolerance = 1e-5
    default_temperature_cache_tolerance = 1e-3
    
    def __init__(self, imol=None, thermal_condition=None, thermo=None,
                 composition_cache_tolerance=None,
                 temperature_cache_tolerance=None,
                 method=None):
        super().__init__(imol, thermal_condition, thermo)
        self.composition_cache_tolerance = (
            self.default_composition_cache_tolerance 
            if composition_cache_tolerance  is None else 
            composition_cache_tolerance 
        )
        self.temperature_cache_tolerance = (
            self.default_temperature_cache_tolerance 
            if temperature_cache_tolerance is None else 
            temperature_cache_tolerance
        )
        self.method = self.default_method if method is None else method
        self._lle_chemicals = None
        self._K = None
        self._phi = None
    
    def __call__(self, T, P=None, top_chemical=None, update=True, use_cache=True, single_loop=False):
        """
        Perform liquid-liquid equilibrium.

        Parameters
        ----------
        T : float
            Operating temperature [K].
        P : float, optional
            Operating pressure [Pa].
        top_chemical : str, optional
            Identifier of chemical that will be favored in the "LIQUID" phase.
        update : bool, optional
            Whether to update material flows, temperature and pressure. If False,
            returns the chemicals in liquid-liquid equilibrium, 
            partition coefficients, and phase fraction.
            
        """
        imol = self._imol
        if update:
            thermal_condition = self._thermal_condition
            thermal_condition.T = T
            if P: thermal_condition.P = P
        else:
            old_data = imol.data.copy()
        mol, index, lle_chemicals = self.get_liquid_mol_data()
        F_mol = mol.sum()
        if F_mol and len(lle_chemicals) > 1:
            z_mol = mol = mol / F_mol # Normalize first
            use_cache = (
                use_cache 
                and self._lle_chemicals == lle_chemicals
                and T - self._T < self.temperature_cache_tolerance 
                and (self._z_mol - z_mol < self.composition_cache_tolerance).all()
            )
            if use_cache:
                K = self._K 
                self._phi = phi = phase_fraction(z_mol, K, self._phi)
                if phi >= 1.:
                    mol_l = mol
                    mol_L = 0. * mol
                else:
                    y = z_mol * K / (phi * K + (1 - phi))
                    mol_l = y * phi
                    mol_L = mol - mol_l
            else:
                if self._lle_chemicals != lle_chemicals: 
                    self._K = None
                    self._gamma_y = None
                    self._phi = None
                mol_L = self.solve_lle_liquid_mol(mol, T, lle_chemicals, single_loop)
                mol_l = mol - mol_L
                if (mol_L < 0).any(): 
                    breakpoint()
                    self.solve_lle_liquid_mol(mol, T, lle_chemicals, single_loop)
                if (mol_l < 0).any(): 
                    breakpoint()
                    self.solve_lle_liquid_mol(mol, T, lle_chemicals, single_loop)
            if top_chemical:
                MW = self.chemicals.MW[index]
                mass_L = mol_L * MW
                mass_l = mol_l * MW
                IDs = {i.ID: n for n, i in enumerate(lle_chemicals)}
                try: top_chemical_index = IDs[top_chemical]
                except: pass
                else:
                    ML = mass_L.sum()
                    Ml = mass_l.sum()
                    if ML and Ml:
                        C_L = mass_L[top_chemical_index] / ML
                        C_l = mass_l[top_chemical_index] / Ml
                        if C_L < C_l: mol_l, mol_L = mol_L, mol_l
                    elif Ml:
                        mol_l, mol_L = mol_L, mol_l
            F_mol_l = mol_l.sum()
            F_mol_L = mol_L.sum()
            if not F_mol_L:
                self._K = np.zeros_like(mol)
                self._gamma_y = np.ones_like(mol)
                self._phi = 0.
            elif not F_mol_l:
                self._K = 1e16 * np.ones_like(mol)
                self._gamma_y = np.ones_like(mol)
                self._phi = 1.
            else:
                x_mol_l = mol_l / F_mol_l
                x_mol_L = mol_L / F_mol_L
                x_mol_l[x_mol_l < 1e-16] = 1e-16
                K = x_mol_L / x_mol_l
                gamma = self.thermo.Gamma(lle_chemicals)
                if (K < 0).any(): breakpoint()
                self._K = K
                self._gamma_y = gamma(x_mol_L, T)
                self._phi = F_mol_L / (F_mol_L + F_mol_l)
            self._lle_chemicals = lle_chemicals
            self._z_mol = z_mol
            self._T = T
            if not update: 
                imol.data[:] = old_data
                return self._lle_chemicals, self._K, self._gamma_y, self._phi
            imol['l'][index] = mol_l * F_mol
            imol['L'][index] = mol_L * F_mol
        elif not update: 
            mol_l = mol
            mol_L = np.zeros_like(mol_l)
            if top_chemical:
                MW = self.chemicals.MW[index]
                IDs = {i.ID: n for n, i in enumerate(lle_chemicals)}
                try: top_chemical_index = IDs[top_chemical]
                except: pass
                else:
                    C_L = mol_L[top_chemical_index]
                    C_l = mol_l[top_chemical_index]
                    if C_L < C_l: mol_l, mol_L = mol_L, mol_l
            F_mol_L = mol_L.sum()
            if F_mol_L:
                K = 1e16 * np.ones_like(mol)
                phi = 1.
            else:
                K = np.zeros_like(mol)
                phi = 0.
            gamma_y = np.ones_like(mol)
            imol.data[:] = old_data
            return lle_chemicals, K, gamma_y, phi
        
    def solve_lle_liquid_mol(self, mol, T, lle_chemicals, single_loop):
        gamma = self.thermo.Gamma(lle_chemicals)
        indices = np.argsort(mol * np.array([i.MW for i in lle_chemicals]))
        method = self.method
        n = mol.size
        if method == 'pseudo equilibrium':
            if self._K is not None and 0 < self._phi < 1:
                K = self._K
                phi = self._phi
            else:
                x = mol.copy()
                y = mol.copy()
                a = indices[-1]
                b = indices[-2]
                x[a] = 0.99
                y[a] = 1e-3
                x[b] = 1e-3
                y[b] = 0.99
                x /= x.sum()
                y /= y.sum()
                K = gamma(y, T) / gamma(x, T)
                phi = 0.5
            if single_loop:
                z = mol
                f_gamma = gamma.f
                gamma_args = gamma.args
                phi = phase_fraction(z, K, phi)
                try:
                    x = z/(1. + phi * (K - 1.))
                except:
                    x = np.ones(n)
                x /= x.sum()
                y = K * x
                Kgammayphi = np.zeros(2*n + 1)
                Kgammayphi[:n] = K
                Kgammayphi[n:-1] = f_gamma(y, T, *gamma_args)
                Kgammayphi[-1] = phi
                Kgammay = Kgammayphi[:-1]
                phi = Kgammayphi[-1]
                args=(z, T, n, f_gamma, gamma_args)
                Kgammay = flx.fixed_point(
                    psuedo_equilibrium_inner_loop, Kgammay, 
                    args=(*args, phi), **self.pseudo_equilibrium_inner_loop_options,
                )
                K = Kgammay[:n]
                try:
                    phi = phase_fraction(z, K, phi)
                except (ZeroDivisionError, FloatingPointError):
                    return z
                if np.isnan(phi): return z
                if phi > 1: phi = 1 - 1e-16
                if phi < 0: phi = 1e-16
                return z/(1. + phi * (K - 1.)) * (1 - phi)
            else:
                K[K == 0] = 1e-9
                if (K <= 0).any(): breakpoint()
                return pseudo_equilibrium(
                    K, phi, mol, T, n, gamma.f, gamma.args, 
                    self.pseudo_equilibrium_inner_loop_options,
                    self.pseudo_equilibrium_outer_loop_options,
                )
        index = indices[-1]
        args = (mol, T, gamma.f, gamma.args)
        bounds = np.zeros([n, 2])
        bounds[:, 1] = mol
        bounds[index, 1] = 0.5 * mol[index] # Remove symmetry
        if method == 'shgo':
            result = shgo(
                lle_objective_function, bounds, args,
                options=self.shgo_options
            )
            if not result.success or (result.x == 0.).all():
                result = differential_evolution(
                    lle_objective_function, bounds, args,
                    **self.differential_evolution_options
                )
            return result.x
        elif method == 'differential evolution':
            result = differential_evolution(
                lle_objective_function, bounds, args,
                **self.differential_evolution_options
            )
            return result.x
        else:
            raise ValueError(f"invalid method {repr(method)}")
        
    def get_liquid_mol_data(self):
        # Get flow rates
        imol = self._imol
        imol['L'] = mol =  imol['l'] + imol['L']
        imol['l'] = 0
        index = self.chemicals.get_lle_indices(mol.nonzero_keys())
        mol = mol[index]
        chemicals = self.chemicals.tuple
        lle_chemicals = [chemicals[i] for i in index]
        return mol, index, lle_chemicals

class LLECache(Cache): load = LLE
del Cache, njit, Equilibrium



# -*- coding: utf-8 -*-
"""
"""
from typing import NamedTuple
from warnings import warn
from graphviz import Digraph, Source
from IPython import display as _display
from collections import deque
from colorpalette import Color
from scipy.interpolate import interp1d
from matplotlib import pyplot as plt
from itertools import product
from PIL import Image
import contextlib
import pydot
import numpy as np
import imageio
import os
from matplotlib import colormaps
from matplotlib.colors import rgb2hex

viridis = colormaps['viridis']

__all__ = (
    'PhenomeNode',
    'EquationNode',
    'VariableNode',   
    'PhenomeGraph',
    'BipartitePhenomeGraph',
    'PhenomenaGraph',
    'DotFile',
    'Edge',
    'all_subgraphs',
)

#: [dict] Default variable node settings
variable_node_options = {
    'gradientangle': '0',
    'width': '0.15',
    'height': '0.15',
    'orientation': '0.0',
    'peripheries': '1',
    'penwidth': '3',
    'margin': '0',
    'fillcolor': 'none',
    'shape': 'circle',
    'style': 'filled',
    # 'color': '#b4b1ae',
    'label': '',
    'pin': 'true',
}

#: [dict] Default equation node settings
equation_node_options = {
    'shape': 'circle',
    'style': 'filled',
    'gradientangle': '0',
    'width': '0.2',
    'height': '0.2',
    'orientation': '0.0',
    'peripheries': '1',
    'margin': '0',
    'label': '',
    'pin': 'true',
}

#: [dict] Default phenomenode settings
phenomenode_options = {
    'shape': 'box',
    'style': 'filled',
    'gradientangle': '0',
    'width': '0.2',
    'orientation': '0.0',
    'peripheries': '0',
    'margin': '0',
    'label': '',
    'pin': 'true',
}

#: [dict] Default phenomenode edge settings
phenomenode_edge_options = dict(
    label='', 
    arrowtail='none', 
    arrowhead='none', 
    minlen='1',
    dir='none',
    penwidth='3',
    style='solid',
    pin='true',
)

#: [dict] Default equation node colors by category
colors = {
    # Used
    'material': '#a280b9', # purple
    'energy': '#ed586f', # red
    'vle': '#60c1cf', # blue
    'lle': '#79bf82', # green
    'phenomena': '#f98f60', # generic, orange
    'reaction': '#dd7440', # orange
    'pressure': '#00a996', # Darker blue 
    'aggregated': '#5c5763', # black
}
edge_options = dict(
    label='', 
    arrowtail='none', 
    arrowhead='none', 
    headport='c', 
    tailport='c', 
    len='0.5',
    # minlen='0.5',
    dir='none',
    penwidth='3',
    style='solid',
    # pin='true',
)
folder = os.path.dirname(__file__)
folder = os.path.join(folder, 'temporary_images')
             
all_subgraphs = ('material_balance', 'energy_balance', 'phenomena')

def filter_nodes(nodes, cls):
    filtered_nodes = []
    past_nodes = set()
    for i in nodes:
        if i is None: continue
        if i in past_nodes: continue    
        if not isinstance(i, cls):
            raise ValueError('equation nodes can only be connected to variable nodes')
        filtered_nodes.append(i)
        past_nodes.add(i)
    return filtered_nodes

def group_equations(equations, criteria, all_equations, all_neighbors):
    for eq in equations:
        if eq in all_equations or eq in all_neighbors: continue
        if criteria.match(eq.name): 
            all_equations.add(eq)
            group_equations(eq.neighbors, criteria, all_equations, all_neighbors)
        else:
            all_neighbors.add(eq)

def get_order(values):
    N = len(values)
    order = [*range(N)]
    order = sorted(order, key=lambda x: values[x])
    past_values = {}
    for n, i in enumerate(order):
        value = values[i]
        if value in past_values:
            order[n] = past_values[value]
        else:
            past_values[value] = i
    return order

class EquationNode:
    __slots__ = ('name', 'inputs', 'outputs')
    
    def __init__(self, name):
        self.name = name    
    
    @property
    def category(self):
        name = self.name
        for category in ('material', 'energy', 'vle', 'lle'):
            if category in name: return category
        return 'phenomena' # generic
    
    def set_equations(self, inputs=(), outputs=()):
        self.inputs = filter_nodes(inputs, VariableNode)
        self.outputs = filter_nodes(outputs, VariableNode)
        for i in self.variables: i.equations.append(self)
        
    @property
    def neighbors(self):
        return set([j for i in self.variables for j in i.equations])
        
    @property
    def variables(self):
        return (*self.inputs, *self.outputs)
    
    def get_edges(self, inputs=True, outputs=True):
        if inputs:
            nodes = self.inputs.copy()
            if outputs:
                nodes += self.outputs
        elif outputs:
            nodes = self.outputs
        else:
            nodes = ()
        return [Edge(self, i) for i in nodes]
        
    def __repr__(self):
        return self.name
    
    
class VariableNode:
    __slots__ = ('name', 'value', 'equations')
    
    def __init__(self, name, value):
        self.name = name
        self.value = value
        self.equations = []
        
    def __repr__(self):
        return self.name


class Edge(NamedTuple):
    equation: EquationNode
    variable: VariableNode
    
    @property
    def name(self):
        return (self.equation.name, self.variable.name)
    

class Criteria:
    __slots__ = ('names',)
    
    def __init__(self, names):
        self.names = names
        
    def match(self, full_name):
        return all([i in full_name for i in self.names])


class CriteriaSelection:
    __slots__ = ('criterias',)
    def __init__(self, *selection):
        self.criterias = [Criteria(i) for i in product(*selection)]

    def get_criteria(self, full_name):
        for criteria in self.criterias:
            if criteria.match(full_name): return criteria
        breakpoint()

class PhenomeNode:
    __slots__ = ('name', 'size', 'equations', 'neighbors', 'criteria')
    
    def __init__(self, equation, criteria_selection):
        self.name = equation.name # Good enough
        self.equations = equations = set([equation])
        self.neighbors = neighbors = set()
        self.criteria = criteria = criteria_selection.get_criteria(equation.name)
        group_equations(equation.neighbors, criteria, equations, neighbors)
        
    category = EquationNode.category
        
    def finalize(self, stage_tags):
        name = '_'.join([i.name for i in self.equations])
        self.size = sum([i in name for i in stage_tags])
        
    def output_variables(self):
        return set([j for i in self.equations for j in i.outputs]) 
        
    def __repr__(self):
        return self.name


class PhenomeGraph:
    __slots__ = ('phenomenodes', 'edges', 'pydot', 'profiles', 'depth')
        
    def __init__(self, equations, criteria_selection, stage_tags, variable_profiles, file=None):
        self.depth = len(criteria_selection) # TODO: remove this workaround for more detailed algorithm
        phenomenodes = []
        collected_equations = set()
        if not isinstance(criteria_selection, CriteriaSelection): criteria_selection = CriteriaSelection(*criteria_selection)
        for eq in equations:
            if eq in collected_equations: continue
            phenomenode = PhenomeNode(eq, criteria_selection)
            phenomenodes.append(phenomenode)
            collected_equations.update(phenomenode.equations)
        reference = {
            eq: node for node in phenomenodes
            for eq in node.equations
        }
        edges = set([
            frozenset([node, reference[eq]]) for node in phenomenodes
            for eq in node.neighbors
        ])
        for i in phenomenodes: i.finalize(stage_tags)
        errors = np.abs(variable_profiles - variable_profiles.iloc[-1])
        profiles = [
            errors[[i.name for i in phenomenode.output_variables()]].values.sum(axis=1)
            for phenomenode in phenomenodes
        ]
        profiles =[get_order(i) for i in profiles]
        # profiles = [
        #     np.diff(
        #         variable_profiles[[i.name for i in phenomenode.output_variables()]].values,
        #         axis=0
        #     ).sum(axis=1)
        #     for phenomenode in phenomenodes
        # ]
        profiles = np.abs(np.array(profiles, dtype=float)).T
        # profiles = np.log(profiles + 1e-6)
        # profiles -= profiles.min(axis=0, keepdims=True)
        max_errors = profiles.max(axis=0, keepdims=True)
        max_errors[max_errors < 1e-9] = 1
        profiles /= max_errors
        self.phenomenodes = phenomenodes
        self.edges = edges
        self.profiles = profiles
        self.load_pydot()
        
    def load_pydot(self, 
            file=None,
            maxiter=None,
            damping=None, 
            K=None,
        ):
        if file:
            try:
                self.pydot = pydot.graph_from_dot_file(file)[0]
            except:
                self.load_pydot(None, maxiter, damping, K)
                self.write(file)
            return
        if maxiter is None: maxiter = '20000'
        if damping is None: damping = '0.3'
        if K is None: K = '0.2'
        # Create a digraph and set direction left to right
        digraph = Digraph(format='png', strict=True)
        digraph.attr(
            'graph',
            rankdir='LR', 
            maxiter=maxiter, 
            Damping=damping, 
            K=K,
            penwidth='0', 
            color='none', 
            bgcolor='transparent',
            nodesep='0.02', 
            ranksep='0.02', 
            layout='dot', 
            splines='curved', 
            outputorder='edgesfirst', 
            dpi='300',
            # nodesep='0.2', 
            # ranksep='0.2',
            overlap='compress', 
            dir='none',
        )
        if self.depth == 1:
            dct = {}
            for i in self.phenomenodes:
                key = i.criteria.names
                if key in dct: dct[key].append(i)
                else: dct[key] = [i]
                
            for n, phenomenodes in enumerate(dct.values()): 
                with digraph.subgraph(name='cluster_' + str(n)) as subgraph:
                    subgraph.attr('graph', color='none', bgcolor='none', shape='box', label='')
                    for i in phenomenodes:
                        color = colors[i.category]
                        subgraph.node(
                            name=i.name, 
                            color='black',
                            fillcolor=color,
                            height=str(i.size * 0.2),
                            **phenomenode_options
                        )
        elif self.depth == 2:
            dct = {}
            for i in self.phenomenodes:
                outer, inner = i.criteria.names
                if outer in dct: subdct = dct[outer]
                else: dct[outer] = subdct = {}
                if inner in subdct: subdct[inner].append(i)
                else: subdct[inner] = [i]
                
            for m, subdct in enumerate(dct.values()):
                with digraph.subgraph(name=f'cluster_{m}') as subgraph:
                    subgraph.attr('graph', color='none', bgcolor='none', shape='box', label='')
                    for n, phenomenodes in enumerate(subdct.values()): 
                        with subgraph.subgraph(name=f'cluster_{m}_{n}') as sub2:
                            sub2.attr('graph', color='none', bgcolor='none', shape='box', label='')
                            for i in phenomenodes:
                                color = colors[i.category]
                                sub2.node(
                                    name=i.name, 
                                    color='black',
                                    fillcolor=color,
                                    height=str(i.size * 0.2),
                                    **phenomenode_options
                                )
                                
        # Set attributes for graph and edges
        digraph.attr(
            'edge', 
            concentrate='true',
            dir='none'
        )
        for left, right in self.edges:
            if right.size == 1:
                color = colors[right.category]
            elif left.size == 1:
                color = colors[right.category]
            else:
                color = 'black'
            digraph.attr('edge', label='', taillabel='', headlabel='')
            digraph.edge(
                left.name, right.name, 
                color='black',
                **phenomenode_edge_options,
            )
        dot = digraph.pipe(format='dot') # Pin nodes
        self.pydot = pydot.graph_from_dot_data(dot.decode('utf-8'))[0]
        
    def display(self): # TODO: this does not work, maybe revert back to graphviz  
        source = Source(self.pydot.create_dot().decode('utf-8'))
        # img = source.pipe('svg')
        # x = _display.SVG(img)
        x = _display.Image(source.pipe(format='png'))
        return _display.display(x)
        
    def write(self, file):
        format = file[file.index('.')+1:]
        method = 'write_' + format
        getattr(self.pydot, method)(file)
        
    # TODO: plots/gifs
    def plot_convergence_profile(self, file=None):
        equation_profiles = self.equation_profiles
        profiles_arr = np.abs(equation_profiles.values)
        equations = sorted(self.equations, key=lambda x: x.category)
        index = [equation_profiles.columns.get_loc(i.name) for i in equations]
        # ps = [*range(0, 101, 1)]
        # percentiles = np.percentile(profiles_arr, ps, axis=0)
        # fs = [interp1d(i, ps) for i in percentiles.T]
        M, N = profiles_arr.shape
        image = np.zeros([M, N, 3])
        # errors_percentiles = np.zeros([M, N])
        # for i in range(M):
        #     for j in index: 
        #         errors_percentiles[i, j] = fs[j](profiles_arr[i, j])
        # errors_percentiles -= errors_percentiles.min(axis=0, keepdims=True)
        profiles_arr[:] = np.log(profiles_arr + 1e-6)
        profiles_arr -= profiles_arr.min(axis=0, keepdims=True)
        max_errors = profiles_arr.max(axis=0, keepdims=True)
        max_errors[max_errors < 1e-9] = 1
        profiles_arr /= max_errors
        for i in range(M):
            for k, (j, eq) in enumerate(zip(index, equations)): 
                # error = errors_percentiles[i, j]
                # if error < 20: error = 0.
                error = 100 * profiles_arr[i, j]
                color = colors[eq.category]
                image[i, k, :] = Color(fg=color).shade(error).RGBn        
        ax = plt.imshow(image)
        plt.axis('off')
        if file:
            for i in ('svg', 'png'):
                plt.savefig(file, dpi=900, transparent=True)
        return ax
        
    def convergence_gif(self, file=None, duration=300, **kwargs):
        digraph = self.pydot
        digraph.set_maxiter('10') # This speeds up image creation.
        digraph.set_dpi('100') # Pydot does not seem to save dpi from piped graphviz
        # breakpoint()
        output_file = file
        input_file = os.path.join(folder, 'temp0.png')
        profiles = self.profiles
        M, N = profiles.shape
        phenomenodes = self.phenomenodes
        nodes_dct = {}
        for subgraph in digraph.get_subgraph_list():
            if self.depth == 1:
                for i in subgraph.get_nodes():
                    nodes_dct[i.get_name().strip('"')] = i
            elif self.depth == 2:
                for subgraph in subgraph.get_subgraph_list():
                    for i in subgraph.get_nodes():
                        nodes_dct[i.get_name().strip('"')] = i
        print('M', M)
        for n in range(M):
            print(n)
            continue
            for j, node in enumerate(phenomenodes): 
                error = 0.8 * profiles[n, j] # Leave at least 20% of color in
                color = Color(fg=colors[node.category]).shade(error).HEX
                node = nodes_dct[node.name]
                node.set_fillcolor(color)
            digraph.write_png(input_file.replace('0', str(n)))
            
        if output_file is None: output_file = os.path.join(folder, 'temp.gif')
        elif '.' not in output_file: output_file += '.gif'
        
        print('duration', duration)
        with contextlib.ExitStack() as stack:
            # lazily load images
            imgs = (
                stack.enter_context(
                    Image.open(
                        input_file.replace('0', str(i))
                    )
                ) for i in range(M)
            )
        
            # extract  first image from iterator
            img = next(imgs)
        
            # https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html#gif
            img.save(fp=output_file, format='GIF', append_images=imgs,
                     save_all=True, duration=duration, optimize=False, loop=1)
            
        

class BipartitePhenomeGraph:
    __slots__ = ('name', 'equations', 'variables', 'edges', 
                 'variable_profiles', 'equation_profiles', 'edge_profiles',
                 'subgraphs', 'pydot')
    
    @staticmethod
    def get_node_edges(equation_nodes, inputs=True, outputs=True):
        return set(sum([i.get_edges(inputs, outputs) for i in equation_nodes], []))
    
    def __init__(self, 
            name, equations, variables, edges, 
            equation_profiles, variable_profiles, 
            edge_profiles, file=None,
            **kwargs,
        ):
        self.name = name
        self.equations = equations
        self.variables = variables
        self.edges = edges
        self.equation_profiles = equation_profiles
        self.variable_profiles = variable_profiles
        self.edge_profiles = edge_profiles
        self.subgraphs = []
        self.load_pydot(file, **kwargs)
       
    def subgraph(self, 
            name, equations, variables, edges, 
            equation_profiles, variable_profiles, 
            edge_profiles, file=None, **kwargs,
        ):
        subgraph = BipartitePhenomeGraph.__new__(BipartitePhenomeGraph)
        subgraph.name = name
        subgraph.equations = equations
        subgraph.variables = variables
        subgraph.edges = edges
        subgraph.equation_profiles = equation_profiles
        subgraph.variable_profiles = variable_profiles
        subgraph.edge_profiles = edge_profiles
        self.subgraphs.append(subgraph)
        subgraph.load_pydot(file, parent=self.pydot, **kwargs)
       
    def load_pydot(self, 
            file=None,
            maxiter=None,
            damping=None, 
            K=None,
            subgraph_units=None,
            parent=None,
        ):
        if file:
            try:
                self.pydot = pydot.graph_from_dot_file(file)[0]
            except:
                self.load_pydot(None, maxiter, damping, K, subgraph_units, parent)
                self.write(file)
            return
        if maxiter is None: maxiter = '10000'
        if damping is None: damping = '0.2'
        if K is None: K = '0.2'
        # Create a digraph and set direction left to right
        digraph = Digraph(format='png', strict=True)
        digraph.attr(
            'graph',
            rankdir='LR', 
            maxiter=maxiter, 
            Damping=damping, 
            K=K,
            sep='1',
            penwidth='0', 
            color='none', 
            bgcolor='transparent',
            nodesep='0.02', 
            ranksep='0.02', 
            layout='fdp', 
            splines='curved', 
            outputorder='edgesfirst', 
            dpi='300',
            # nodesep='0.2', 
            # ranksep='0.2',
            overlap='compress', 
            dir='none',
        )
        variables = self.variables
        equations = self.equations
        edges = self.edges
        inner_edges = []
        past_nodes = set()
        subgraph_units = False
        if subgraph_units:
            for name in subgraph_units:
                # with digraph.subgraph(name=name) as subgraph:
                # subgraph.attr('graph', color='#b4b1ae', bgcolor='#b4b1ae', shape='box', layout='fdp')
                subgraph_equations = [i for i in equations if name in i.name]
                subgraph_variables = set(sum([i.outputs for i in subgraph_equations], []))
                subgraph_variables = set([i for i in subgraph_variables if name in i.name])
                assert subgraph_variables.isdisjoint(past_nodes)
                past_nodes.update(subgraph_variables)
                subgraph_edges = BipartitePhenomeGraph.get_node_edges(subgraph_equations)
                subgraph_edges = tuple([i for i in subgraph_edges if i.variable in subgraph_variables])
                subgraph_variables = tuple(subgraph_variables)
                inner_edges.extend(subgraph_edges)
                self.fill_digraph_nodes(digraph, subgraph_variables, subgraph_equations)
                self.fill_digraph_edges(digraph, subgraph_edges)
            self.fill_digraph_edges(digraph, set(edges).difference(inner_edges))
        else:
            self.fill_digraph_nodes(digraph, variables, equations)
            self.fill_digraph_edges(digraph, edges)
        # img = digraph.pipe(format='png') # Pin nodes
        # x = _display.Image(img)
        # _display.display(x)
        dot = digraph.pipe(format='dot') # Pin nodes
        self.pydot = pydot.graph_from_dot_data(dot.decode('utf-8'))[0]
        # for node in self.pydot.get_nodes(): node.set_pin('false')
        if parent:
            nodes_dct = {i.get_name().strip('"'): i for i in self.pydot.get_nodes()}
            parent_nodes_dct = {i.get_name().strip('"'): i for i in parent.get_nodes()}
            edges_dct = {
                (i.get_source().replace(':c', '').strip('"'), i.get_destination().replace(':c', '').strip('"')): i
                for i in self.pydot.get_edges()
            }
            parent_edges_dct = {
                (i.get_source().replace(':c', '').strip('"'), i.get_destination().replace(':c', '').strip('"')): i
                for i in parent.get_edges()
            }
            for name, node in nodes_dct.items():
                pos = parent_nodes_dct[name].get_pos()
                if pos: node.set_pos(pos)
            for key, edge in edges_dct.items():
                pos = parent_edges_dct[key].get_pos()
                if pos: edge.set_pos(pos)
        
    def fill_digraph_nodes(self, digraph, variables, equations):
        variable_colors = {}
        for i in equations: 
            color = colors[i.category]
            digraph.node(
                name=i.name, 
                color=color,
                fillcolor=color,
                **equation_node_options
            )
            for var in i.outputs: variable_colors[var.name] = color
        for i in variables: 
            digraph.node(
                name=i.name, 
                color=variable_colors[i.name],
                **variable_node_options
            )
    
    def fill_digraph_edges(self, digraph, edges):
        # Set attributes for graph and edges
        digraph.attr(
            'edge', 
            dir='none'
        )
        equations = set(self.equations)
        variables = set(self.variables)
        for edge in edges:
            equation = edge.equation
            variable = edge.variable
            assert equation in equations
            assert variable in variables
            digraph.attr('edge', label='', taillabel='', headlabel='')
            color = colors[equation.category]
            digraph.edge(
                equation.name, variable.name, 
                color=color, name='_'.join(edge.name),
                **edge_options,
            )
        
    def display(self): # TODO: this does not work, maybe revert back to graphviz  
        source = Source(self.pydot.create_dot().decode('utf-8'))
        # img = source.pipe('svg')
        # x = _display.SVG(img)
        x = _display.Image(source.pipe(format='png'))
        return _display.display(x)
        
    def write(self, file):
        format = file[file.index('.')+1:]
        method = 'write_' + format
        getattr(self.pydot, method)(file)
        
    def plot_convergence_profile(self, file=None):
        equation_profiles = self.equation_profiles
        profiles_arr = np.abs(equation_profiles.values)
        equations = sorted(self.equations, key=lambda x: x.category)
        index = [equation_profiles.columns.get_loc(i.name) for i in equations]
        # ps = [*range(0, 101, 1)]
        # percentiles = np.percentile(profiles_arr, ps, axis=0)
        # fs = [interp1d(i, ps) for i in percentiles.T]
        M, N = profiles_arr.shape
        image = np.zeros([M, N, 3])
        # errors_percentiles = np.zeros([M, N])
        # for i in range(M):
        #     for j in index: 
        #         errors_percentiles[i, j] = fs[j](profiles_arr[i, j])
        # errors_percentiles -= errors_percentiles.min(axis=0, keepdims=True)
        profiles_arr[:] = np.log(profiles_arr + 1e-6)
        profiles_arr -= profiles_arr.min(axis=0, keepdims=True)
        max_errors = profiles_arr.max(axis=0, keepdims=True)
        max_errors[max_errors < 1e-9] = 1
        profiles_arr /= max_errors
        for i in range(M):
            for k, (j, eq) in enumerate(zip(index, equations)): 
                # error = errors_percentiles[i, j]
                # if error < 20: error = 0.
                error = 100 * profiles_arr[i, j]
                color = colors[eq.category]
                image[i, k, :] = Color(fg=color).shade(error).RGBn        
        ax = plt.imshow(image)
        plt.axis('off')
        if file:
            for i in ('svg', 'png'):
                plt.savefig(file, dpi=900, transparent=True)
        return ax
        
    def convergence_gif(self, 
            file=None, duration=300, eqp=None, varp=None, 
            reference=None, profiles=None, categorize=True,
            inverse=True, **kwargs
        ):
        digraph = self.pydot
        equations = self.equations
        digraph.set_maxiter('10') # This speeds up image creation.
        digraph.set_dpi('100') # Pydot does not seem to save dpi from piped graphviz
        # breakpoint()
        output_file = file
        input_file = os.path.join(folder, 'temp0.png')
        if reference is None: 
            if eqp is None: eqp = True
        else:
            if eqp is None: eqp = False
        if varp is None: varp = False
        if reference:
            M, N = profiles.shape
        if eqp:
            equation_profiles = self.equation_profiles.copy()
            equation_profiles_arr = equation_profiles.values
            index = [equation_profiles.columns.get_loc(i.name) for i in equations]
            # equation_profiles_arr[:] = np.log(equation_profiles_arr + 1e-6)
            equation_profiles_arr -= equation_profiles_arr.min(axis=0, keepdims=True)
            max_errors = equation_profiles_arr.max(axis=0, keepdims=True)
            max_errors[max_errors < 1e-9] = 1
            equation_profiles_arr /= max_errors
            M, N = equation_profiles_arr.shape
        if varp:
            variable_profiles = self.variable_profiles
            # variable_profiles_arr = np.abs(np.diff(variable_profiles.values, axis=0))
            # variable_profiles_arr = np.log(np.abs(variable_profiles.values - variable_profiles.values[-1]) + 1e-12)
            variable_profiles_arr = np.abs(variable_profiles.values - variable_profiles.values[-1])
            variable_profiles_arr -= variable_profiles_arr.min(axis=0, keepdims=True)
            max_errors = variable_profiles_arr.max(axis=0, keepdims=True)
            max_errors[max_errors < 1e-9] = 1
            variable_profiles_arr /= max_errors
            M, N = variable_profiles_arr.shape
        variables = set(self.variables)
        nodes_dct = {i.get_name().strip('"'): i for i in digraph.get_nodes()}
        edges_dct = {
            (i.get_source().replace(':c', '').strip('"'), i.get_destination().replace(':c', '').strip('"')): i
            for i in digraph.get_edges()
        }
        if categorize:
            variable_colors = {}
            for eq in self.equations: 
                color = colors[eq.category]
                for var in eq.outputs: variable_colors[var.name] = color
        
        print('M', M)
        add = 0
        profiles = 100 * profiles
        for n in range(M):
            print(n)
            if reference:
                for eq in equations:
                    index = reference[eq.name]
                    error = profiles[n, index]
                    if inverse:
                        error = 100 - error
                    if categorize:
                        if not inverse: 0.8 * error
                        color = colors[eq.category]
                        color = Color(fg=color).shade(error).HEX
                    else:
                        color = rgb2hex(viridis(profiles[n, index]))
                    node = nodes_dct[eq.name]
                    node.set_color(color)
                    node.set_fillcolor(color)
                    edges = [i for i in eq.get_edges() if i.variable in variables]
                    for edge in edges:
                        _edge = edges_dct[edge.equation.name, edge.variable.name]
                        _edge.set_color(color)
                        var = nodes_dct[edge.variable.name]
                        var.set_color(color)
                digraph.write_png(input_file.replace('0', str(n + add)))
                if n == 0: continue
                add += 1
                for eq in equations:
                    index = reference[eq.name]
                    last = profiles[n - 1, index]
                    baseline = profiles[n, index]
                    error = max((2 * baseline - last), 0) # Leave at least 20% of color in
                    error = min(error, baseline) # Leave at least 20% of color in
                    if inverse:
                        error = 100 - error
                    if categorize:
                        if not inverse: 0.8 * error
                        color = colors[eq.category]
                        color = Color(fg=color).shade(error).HEX
                    else:
                        color = rgb2hex(viridis(profiles[n, index]))
                    node = nodes_dct[eq.name]
                    node.set_color(color)
                    node.set_fillcolor(color)
                    edges = [i for i in eq.get_edges() if i.variable in variables]
                    for edge in edges:
                        _edge = edges_dct[edge.equation.name, edge.variable.name]
                        _edge.set_color(color)
                        var = nodes_dct[edge.variable.name]
                        var.set_color(color)
                digraph.write_png(input_file.replace('0', str(n + add)))
                add += 1
                for eq in equations:
                    index = reference[eq.name]
                    error = profiles[n, index] 
                    if inverse:
                        error = 100 - error
                    if categorize:
                        if not inverse: 0.8 * error
                        color = colors[eq.category]
                        color = Color(fg=color).shade(error).HEX
                    else:
                        color = rgb2hex(viridis(profiles[n, index]))
                    node = nodes_dct[eq.name]
                    node.set_color(color)
                    node.set_fillcolor(color)
                    edges = [i for i in eq.get_edges() if i.variable in variables]
                    for edge in edges:
                        _edge = edges_dct[edge.equation.name, edge.variable.name]
                        _edge.set_color(color)
                        var = nodes_dct[edge.variable.name]
                        var.set_color(color)
            else:
                if eqp:
                    for j, eq in zip(index, equations): 
                        color = colors[eq.category]
                        error = 0.8 * equation_profiles_arr[n, j] # Leave at least 20% of color in
                        color = Color(fg=color).shade(error).HEX
                        node = nodes_dct[eq.name]
                        node.set_color(color)
                        node.set_fillcolor(color)
                        edges = [i for i in eq.get_edges() if i.variable in variables]
                        for edge in edges:
                            _edge = edges_dct[edge.equation.name, edge.variable.name]
                            _edge.set_color(color)
                            if varp: continue
                            var = nodes_dct[edge.variable.name]
                            var.set_color(color)
                if varp:
                    for j, var in enumerate(variables): 
                        var_color = Color(fg=variable_colors[var.name])
                        error = 0.8 * variable_profiles_arr[n, j] # Leave at least 20% of color in
                        color = var_color.shade(error).HEX
                        node = nodes_dct[var.name]
                        node.set_color(color)
            digraph.write_png(input_file.replace('0', str(n + add)))
            
        if output_file is None: output_file = os.path.join(folder, 'temp.gif')
        elif '.' not in output_file: output_file += '.gif'
        
        print('duration', duration)
        with contextlib.ExitStack() as stack:
            # lazily load images
            imgs = (
                stack.enter_context(
                    Image.open(
                        input_file.replace('0', str(i))
                    )
                ) for i in range(M + add)
            )
        
            # extract  first image from iterator
            img = next(imgs)
        
            # https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html#gif
            img.save(fp=output_file, format='GIF', append_images=imgs,
                     save_all=True, duration=duration, optimize=False, loop=1)
            
    def __repr__(self):
        return self.name
    
PhenomenaGraph = BipartitePhenomeGraph # For backwards compatibility

class DotFile:
    __slots__ = ('file', 'subfiles')
    
    def __init__(self, file):
        self.file = file if file.endswith('.dot') else f'{file}.dot'
         
    def subfile(self, subgraph):
        return self.file.replace('.dot', f'_{subgraph}.dot')
        
    def __repr__(self):
        return f"{type(self).__name__}(file={self.file!r})"


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from __future__ import annotations
from ._thermo import Thermo
from ._stream import Stream
from ._thermal_condition import ThermalCondition
from .indexer import MolarFlowIndexer, nonzeros, parent_indexer
from ._phase import phase_tuple
from . import equilibrium as eq
from . import utils
import numpy as np
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from numpy.typing import NDArray
    from typing import Optional, Sequence, Dict, Tuple

__all__ = ('MultiStream', )

def get_phase_fraction(stream, phases):
    all_phases = stream.phases
    phase_fraction = 0.
    F_mol = stream.F_mol
    imol = stream.imol
    if not F_mol: return 0.
    for phase in phases:
        if phase in all_phases:
            phase_fraction += imol[phase].sum() 
    return phase_fraction / F_mol

@utils.registered_franchise(Stream)
class MultiStream(Stream):
    """
    Create a MultiStream object that defines material flow rates for multiple
    phases along with its thermodynamic state. Thermodynamic and transport
    properties of a stream are available as properties, while thermodynamic
    equilbrium (e.g. VLE, and bubble and dew points) are available as methods. 

    Parameters
    ----------
    ID : 
        A unique identification. If ID is None, stream will not be registered.
        If no ID is given, stream will be registered with a unique ID.
    flow : 
        All flow rates corresponding to `phases` by row and chemical IDs by column.
    T : 
        Temperature [K]. Defaults to 298.15.
    P : 
        Pressure [Pa]. Defaults to 101325.
    phases : 
        Tuple denoting the phases present. Defaults to ('g', 'l').
    units : 
        Flow rate units of measure (only mass, molar, and
        volumetric flow rates are valid). Defaults to 'kmol/hr'.
    price : 
        Price per unit mass [USD/kg]. Defaults to 0.
    total_flow : 
        Total flow rate.
    thermo : 
        Thermodynamic equilibrium package. Defaults to `thermosteam.settings.get_thermo()`.
    vlle :
        Whether to run rigorous phase equilibrium to determine phases. 
        Defaults to False.
    **phase_flow : tuple[str, float]
        phase-(ID, flow) pairs.
    
    Examples
    --------
    Before creating streams, first set the chemicals:
        
    >>> import thermosteam as tmo
    >>> tmo.settings.set_thermo(['Water', 'Ethanol'], cache=True)
    
    Create a multi phase stream, defining the thermodynamic condition and
    flow rates:
        
    >>> s1 = tmo.MultiStream(ID='s1',T=298.15, P=101325,
    ...                      l=[('Water', 20), ('Ethanol', 10)], units='kg/hr')
    >>> s1.show(flow='kg/hr') # Use the show method to select units of display
    MultiStream: s1
    phases: ('g', 'l'), T: 298.15 K, P: 101325 Pa
    flow (kg/hr): (l) Water    20
                      Ethanol  10
    
    The temperature and pressure are stored as attributes:
    
    >>> (s1.T, s1.P)
    (298.15, 101325.0)
    
    Unlike Stream objects, the `mol` attribute does not store data, it simply
    returns the total flow rate of each chemical. Setting an element of the
    array raises an error to prevent the wrong assumption that the data is
    linked:
    
    >>> s1.mol
    sparse([1.11 , 0.217])
    >>> s1.mol[0] = 1
    Traceback (most recent call last):
    ValueError: assignment destination is read-only
    
    All flow rates are stored in the `imol` attribute:
    
    >>> s1.imol.show() # Molar flow rates [kmol/hr]
    MolarFlowIndexer (kmol/hr):
     (l) Water     1.11
         Ethanol   0.2171
    >>> # Index a single chemical in the liquid phase
    >>> s1.imol['l', 'Water']
    1.1101
    >>> # Index multiple chemicals in the liquid phase
    >>> s1.imol['l', ('Ethanol', 'Water')]
    array([0.217, 1.11 ])
    >>> # Index the vapor phase
    >>> s1.imol['g']
    sparse([0., 0.])
    >>> # Index flow of chemicals summed across all phases
    >>> s1.imol['Ethanol', 'Water']
    array([0.217, 1.11 ])
    
    Note that overall chemical flows in MultiStream objects cannot be set like
    with Stream objects:
        
    >>> # s1.imol['Ethanol', 'Water'] = [1, 0]
    Traceback (most recent call last):
    IndexError: multiple phases present; must include phase key to set chemical data
    
    Chemical flows must be set by phase:

    >>> s1.imol['l', ('Ethanol', 'Water')] = [1, 0]
    
    The most convinient way to get and set flow rates is through the 
    `get_flow` and `set_flow` methods:
    
    >>> # Set flow
    >>> key = ('l', 'Water')
    >>> s1.set_flow(1, 'gpm', key)
    >>> s1.get_flow('gpm', key)
    1.0
    >>> # Set multiple flows
    >>> key = ('l', ('Ethanol', 'Water'))
    >>> s1.set_flow([10, 20], 'kg/hr', key)
    >>> s1.get_flow('kg/hr', key)
    array([10., 20.])
    
    Chemical flows across all phases can be retrieved if no phase is
    given:
        
    >>> s1.get_flow('kg/hr', ('Ethanol', 'Water'))
    array([10., 20.])
    
    However, setting chemical data requires the phase to be specified:
        
    >>> s1.set_flow([10, 20], 'kg/hr', ('Ethanol', 'Water'))
    Traceback (most recent call last):
    IndexError: multiple phases present; must include phase key to set chemical data
    
    Note that for both Stream and MultiStream objects, `mol`, `imol`, 
    and `get_flow` return chemical flows across all phases when given only 
    chemical IDs.
    
    Vapor-liquid equilibrium can be performed by setting 2 degrees of freedom
    from the following list:
    
    * T - Temperature [K]
    * P - Pressure [Pa]
    * V - Vapor fraction
    * H - Enthalpy [kJ/hr]
        
    >>> s1.vle(P=101325, T=365)
    
    Each phase can be accessed separately too:
    
    >>> s1['l'].show()
    Stream: 
    phase: 'l', T: 365 K, P: 101325 Pa
    flow (kmol/hr): Water    0.617
                    Ethanol  0.0238
    >>> s1['g'].show()
    Stream: 
    phase: 'g', T: 365 K, P: 101325 Pa
    flow (kmol/hr): Water    0.493
                    Ethanol  0.193
    
    Note that the phase cannot be changed:
    
    >>> s1['g'].phase = 'l'
    Traceback (most recent call last):
    AttributeError: phase is locked
    
    """
    __slots__ = ()
    def __init__(self, 
                 ID: Optional[str]="",
                 flow: Optional[Sequence]=(), 
                 T: Optional[float]=298.15, 
                 P: Optional[float]=101325.,
                 phases: Optional[Sequence[str]]=None, 
                 units: Optional[str]=None, 
                 price: Optional[float]=0,
                 total_flow: Optional[float]=None, 
                 thermo: Optional[Thermo]=None, 
                 characterization_factors: Optional[Dict[str, float]]=None, 
                 vlle: Optional[bool]=False, 
                 **phase_flows: Tuple[str, float]):
        self.characterization_factors = {} if characterization_factors is None else {}
        self._thermal_condition = ThermalCondition(T, P)
        thermo = self._load_thermo(thermo)
        chemicals = thermo.chemicals
        self.price = price
        phases = set(phase_flows).union(['l', 'g']) if phases is None else phases
        if units:
            name, factor = self._get_flow_name_and_factor(units)
            if name == 'mass':
                group_wt_compositions = chemicals._group_wt_compositions
                for phase, chemical_flows in phase_flows.items():
                    new_chemical_flows = []
                    for item in chemical_flows:
                        cID, value = item
                        if cID in group_wt_compositions:
                            compositions = group_wt_compositions[cID]
                            chemical_group = chemicals[cID]
                            for i in range(len(chemical_group)):
                                item = (chemical_group[i]._ID, value * compositions[i])
                                new_chemical_flows.append(item)
                        else:
                            new_chemical_flows.append(item)
                    phase_flows[phase] = new_chemical_flows
            elif name == 'vol':
                group_wt_compositions = chemicals._group_wt_compositions
                for chemical_flows in phase_flows.values():
                    for item in chemical_flows:
                        cID, value = item
                        if cID in group_wt_compositions:
                            raise ValueError(f"cannot set volumetric flow by chemical group '{i}'")
            self._init_indexer(flow, phases, chemicals, phase_flows)
            flow = getattr(self, 'i' + name)
            material_data = self._imol.data / factor
            if total_flow: material_data *= total_flow / material_data.sum()
            flow.data[:] = material_data
        else:
            self._init_indexer(flow, phases, chemicals, phase_flows)
            if total_flow: self._imol.data *= total_flow / self.F_mol
        self._sink = self._source = None
        self.reset_cache()
        self._register(ID)
        if vlle: self.vlle(T, P)
        
    @classmethod
    def from_streams(cls, streams, thermo=None):
        if not streams: raise ValueError('at least one stream must be passed')
        self = cls.__new__(cls)
        self._streams = streams_by_phase = {i.phase: i for i in streams}
        phases = phase_tuple(streams_by_phase)
        N_streams = len(streams)
        if len(phases) != N_streams: raise ValueError('each stream must have a different phase')
        base, *others = streams
        self.characterization_factors = {}
        self._thermal_condition = base._thermal_condition
        for i in others: i._thermal_condition = base._thermal_condition
        self._load_thermo(thermo or base.thermo)
        self.price = 0
        self._imol = MolarFlowIndexer.from_data(
            [streams_by_phase[i]._imol.data for i in phases], phases, 
            chemicals=base.chemicals
        )
        self._sink = self._source = None
        self.reset_cache()
        self._register(None)
        return self
        
    def reset_flow(self, total_flow=None, units=None, phases=None, **phase_flows):
        """
        Convinience method for resetting flow rate data.
        
        Examples
        --------
        >>> import thermosteam as tmo
        >>> tmo.settings.set_thermo(['Water', 'Ethanol'], cache=True)
        >>> s1 = tmo.MultiStream('s1', l=[('Water', 1)])
        >>> s1.reset_flow(g=[('Ethanol', 1)], phases='lgs', units='kg/hr', total_flow=2)
        >>> s1.show('cwt')
        MultiStream: s1
        phases: ('g', 'l', 's'), T: 298.15 K, P: 101325 Pa
        composition (%): (g) Ethanol  100
                             -------  2 kg/hr
        
        """
        imol = self._imol
        imol.empty()
        self.phases = set(phase_flows).union(['l', 'g']) if phases is None else phases
        if phase_flows:
            for phase, data in phase_flows.items():
                keys, values = zip(*data)
                self.set_flow(values, units, (phase, keys))
        if total_flow:
            self.set_total_flow(total_flow, units)
        
    def _init_indexer(self, flow, phases, chemicals, phase_flows):
        if flow == ():
            parent = parent_indexer(chemicals)
            if phase_flows:
                imol = MolarFlowIndexer(phases, chemicals=chemicals, parent=parent, **phase_flows)
            else:
                imol = MolarFlowIndexer.blank(phases, chemicals, parent)
        else:
            assert not phase_flows, ("may specify either 'flow' or "
                                    "'phase_flows', but not both")
            if isinstance(flow, MolarFlowIndexer):
                imol = flow
            else:
                parent = parent_indexer(chemicals)
                imol = parent.to_material_indexer(phases)
                imol.data[:] = flow
        self._imol = imol
    
    def _get_property(self, name, flow=False, nophase=False):
        property_cache = self._property_cache
        thermal_condition = self._thermal_condition
        imol = self._imol
        data = imol.data
        total = data.sum()
        if total == 0.: 
            return 0. if flow else None
        else:
            composition = data / total
            composition_key = [i.dct for i in composition.rows]
            if nophase:
                literal = (thermal_condition._T, thermal_condition._P)
            else:
                literal = (imol._phases, thermal_condition._T, thermal_condition._P)
            last_literal, last_composition_key = self._property_cache_key
            if literal == last_literal and (composition_key == last_composition_key):
                if name in property_cache:
                    value = property_cache.get(name)
                    return value * total if flow else value
            else:
                property_cache.clear()
            self._property_cache_key = (literal, [i.copy() for i in composition_key])
            if nophase:
                calculate = getattr(self.mixture, name)
                self._property_cache[name] = value = calculate(
                    composition.sum(axis=0), *self.thermal_condition
                )
            else:
                calculate = getattr(self.mixture, 'x' + name)
                self._property_cache[name] = value = calculate(
                    zip(imol._phases, composition), *self.thermal_condition
                )
            return value * total if flow else value
    
    def reset_cache(self):
        """Reset cache regarding equilibrium methods."""
        super().reset_cache()
        if hasattr(self, '_streams'): 
            for i in self._streams.values(): i.reset_cache()
        else:
            self._streams = {}
        self._vle_cache = eq.VLECache(self._imol,
                                      self._thermal_condition, 
                                      self._thermo)
        self._lle_cache = eq.LLECache(self._imol,
                                      self._thermal_condition, 
                                      self._thermo)
        self._sle_cache = eq.SLECache(self._imol,
                                      self._thermal_condition, 
                                      self._thermo)
        
    def __getitem__(self, phase):
        streams = self._streams
        if not (isinstance(phase, str) and len(phase) == 1):
            raise ValueError(f'invalid phase index {phase!r}')
        if phase in streams:
            stream = streams[phase]
        else:
            stream = Stream.__new__(Stream)
            stream._ID = stream._sink = stream._source = None
            stream._imol = self._imol.get_phase(phase, lock=True)
            stream._thermal_condition = self._thermal_condition
            stream._thermo = self._thermo
            stream._property_cache = {}
            stream.characterization_factors = {}
            stream._property_cache_key = None, None
            streams[phase] = stream
        return stream
    
    def __iter__(self):
        for i in self.phases: yield self[i]
    
    def __len__(self):
        return len(self.phases)
    
    ### Property getters ###
    
    def get_flow(self, units: str, key=...):
        """
        Return an array of flow rates in requested units.
        
        Parameters
        ----------
        units : 
            Units of measure.
        key : 
            Index key as (phase, IDs), phase, or IDs where:
            
            * phase is a str, ellipsis, or missing.
            * IDs is a str, Sequence[str], ellipisis, or missing.

        Examples
        --------
        >>> import thermosteam as tmo
        >>> tmo.settings.set_thermo(['Water', 'Ethanol'], cache=True)
        >>> s1 = tmo.MultiStream('s1', l=[('Water', 20), ('Ethanol', 10)], units='kg/hr')
        >>> s1.get_flow('kg/hr', ('l', 'Water'))
        20.0
        
        >>> s1.get_flow('kg/hr')
        sparse([20., 10.])

        """
        name, factor = self._get_flow_name_and_factor(units)
        indexer = getattr(self, 'i' + name)
        return factor * indexer[key]
    
    def set_flow(self, data: NDArray[float]|float, units: str, key=...):
        """
        Set flow rates in given units.

        Parameters
        ----------
        data : 
            Flow rate data.
        units :
            Units of measure.
        key : 
            Index key as (phase, IDs), phase, or IDs where:
            
            * phase is a str, ellipsis, or missing.
            * IDs is a str, Sequence[str], ellipisis, or missing.

        Examples
        --------
        >>> import thermosteam as tmo
        >>> tmo.settings.set_thermo(['Water', 'Ethanol'], cache=True)
        >>> s1 = tmo.MultiStream('s1', l=[('Water', 20), ('Ethanol', 10)], units='kg/hr')
        >>> s1.set_flow(10, 'kg/hr', ('l', 'Water'))
        >>> s1.get_flow('kg/hr', ('l', 'Water'))
        10.0

        """
        name, factor = self._get_flow_name_and_factor(units)
        indexer = getattr(self, 'i' + name)
        indexer[key] = np.asarray(data, dtype=float) / factor
    
    ### Stream data ###
    
    @property
    def phases(self) -> tuple[str, ...]:
        """All phases avaiable."""
        return self._imol._phases
    @phases.setter
    def phases(self, phases):
        phases = set(phases)
        if len(phases) == 1:
            self.phase, = phases
        phases = phase_tuple(phases)
        if phases != self.phases:
            self._imol = self._imol.to_material_indexer(phases)
    
    ### Flow properties ###
            
    @property
    def mol(self) -> NDArray[float]:
        """Chemical molar flow rates (total of all phases)."""
        mol = self._imol.data.sum(0)
        mol.read_only = True
        return mol
    @property
    def mass(self) -> NDArray[float]:
        """Chemical mass flow rates (total of all phases)."""
        mass = self.mol * self.chemicals.MW
        mass.read_only = True
        return mass
    @property
    def vol(self) -> NDArray[float]:
        """Chemical volumetric flow rates (total of all phases)."""
        vol = self.ivol.data.sum(0)
        vol.read_only = True
        return vol
    
    ### Net flow properties ###
    
    @property
    def H(self) -> float:
        """Enthalpy flow rate [kJ/hr]."""
        return self._get_property('H', True)
    @H.setter
    def H(self, H):
        if not H and self.isempty(): return
        self.T = self.mixture.xsolve_T_at_HP(
            self._imol, H, *self._thermal_condition
        )

    @property
    def h(self) -> float:
        """Specific enthalpy in kJ/kmol."""
        return self._get_property('H')
    @h.setter
    def h(self, h: float):
        if not h and self.isempty(): return
        self.T = self.mixture.xsolve_T_at_HP(
            self._imol.iter_composition(), h, *self._thermal_condition
        )

    @property
    def S(self) -> float:
        """Absolute entropy flow rate [kJ/hr]."""
        return self._get_property('S', True)
    @S.setter
    def S(self, S):
        if not S and self.isempty(): return
        self.T = self.mixture.xsolve_T_at_SP(
            self._imol, S, *self._thermal_condition
        )
    
    ### Composition properties ###
    
    @property
    def vapor_fraction(self) -> float:
        """Molar vapor fraction."""
        return self.imol['g'].sum() / F_mol if 'g' in self.phases and (F_mol:=self.F_mol) != 0. else 0.
        
    @property
    def liquid_fraction(self) -> float:
        """Molar liquid fraction."""
        return get_phase_fraction(self, 'lL')
    @property
    def solid_fraction(self) -> float:
        """Molar solid fraction."""
        return get_phase_fraction(self, 'sS')
    
    ### Methods ###
    
    def split_to(self, s1, s2, split, energy_balance=True):
        """
        Split molar flow rate from this stream to two others given
        the split fraction or an array of split fractions.
        
        Examples
        --------
        >>> import thermosteam as tmo
        >>> chemicals = tmo.Chemicals(['Water', 'Ethanol'], cache=True)
        >>> tmo.settings.set_thermo(chemicals)
        >>> s = tmo.Stream('s', Water=20, Ethanol=10, units='kg/hr')
        >>> s.phases = ('l', 'g')
        >>> s1 = tmo.Stream('s1')
        >>> s2 = tmo.Stream('s2')
        >>> split = chemicals.kwarray(dict(Water=0.5, Ethanol=0.1))
        >>> s.split_to(s1, s2, split)
        >>> s1.show(flow='kg/hr')
        MultiStream: s1
        phases: ('g', 'l'), T: 298.15 K, P: 101325 Pa
        flow (kg/hr): (l) Water    10
                          Ethanol  1
        
        >>> s2.show(flow='kg/hr')
        MultiStream: s2
        phases: ('g', 'l'), T: 298.15 K, P: 101325 Pa
        flow (kg/hr): (l) Water    10
                          Ethanol  9
        
        
        """
        phases = self.phases
        if energy_balance or isinstance(s1, MultiStream) or isinstance(s2, MultiStream):
            s1.phases = phases
            s2.phases = phases
            for phase in phases: self[phase].split_to(s1[phase], s2[phase], split)
        else:
            Stream.split_to(self, s1, s2, split)
            return
        if energy_balance:
            tc1 = s1._thermal_condition
            tc2 = s2._thermal_condition
            tc = self._thermal_condition
            tc1._T = tc2._T = tc._T
            tc1._P = tc2._P = tc._P
    
    def copy_like(self, other):
        """
        Copy all conditions of another stream.

        Examples
        --------
        Copy data from another stream with the same property package:
        
        >>> import thermosteam as tmo
        >>> tmo.settings.set_thermo(['Water', 'Ethanol'], cache=True)
        >>> s1 = tmo.MultiStream('s1', l=[('Water', 20), ('Ethanol', 10)], units='kg/hr')
        >>> s2 = tmo.Stream('s2', Water=2, units='kg/hr')
        >>> s1.copy_like(s2)
        >>> s1.show(flow='kg/hr')
        MultiStream: s1
        phases: ('g', 'l'), T: 298.15 K, P: 101325 Pa
        flow (kg/hr): (l) Water  2
         
        Copy data from another stream with a different property package:
        
        >>> import thermosteam as tmo
        >>> tmo.settings.set_thermo(['Water', 'Ethanol'], cache=True)
        >>> s1 = tmo.MultiStream('s1', l=[('Water', 20), ('Ethanol', 10)], units='kg/hr')
        >>> tmo.settings.set_thermo(['Water'], cache=True)
        >>> s2 = tmo.Stream('s2', Water=2, units='kg/hr')
        >>> s1.copy_like(s2)
        >>> s1.show(flow='kg/hr')
        MultiStream: s1
        phases: ('g', 'l'), T: 298.15 K, P: 101325 Pa
        flow (kg/hr): (l) Water  2

        """
        self._imol.copy_like(other._imol)
        self._thermal_condition.copy_like(other._thermal_condition)
    
    def copy_flow(self, 
                  other: Stream, 
                  phase: Optional[Sequence[str]|str]=...,
                  IDs: Optional[Sequence[str]|str]=..., *,
                  remove: Optional[bool]=False,
                  exclude: Optional[bool]=False):
        """
        Copy flow rates of another stream to self.
        
        Parameters
        ----------
        other : 
            Flow rates will be copied from here.
        IDs : 
            Chemical IDs. Defaults to all chemicals.
        remove : 
            If True, copied chemicals will be removed from `stream`.
        exclude :
            If True, exclude designated chemicals when copying.
        
        Notes
        -----
        Works just like <Stream>.copy_flow, but the phase must be specified.
        
        Examples
        --------
        Initialize streams:
        
        >>> import thermosteam as tmo
        >>> tmo.settings.set_thermo(['Water', 'Ethanol'], cache=True) 
        >>> s1 = tmo.MultiStream('s1', l=[('Water', 20), ('Ethanol', 10)], units='kg/hr')
        >>> s2 = tmo.MultiStream('s2')
        
        Copy all flows:
        
        >>> s2.copy_flow(s1)
        >>> s2.show(flow='kg/hr')
        MultiStream: s2
        phases: ('g', 'l'), T: 298.15 K, P: 101325 Pa
        flow (kg/hr): (l) Water    20
                          Ethanol  10
        
        Reset and copy just water flow:
        
        >>> s2.empty()
        >>> s2.copy_flow(s1, IDs='Water')
        >>> s2.show(flow='kg/hr')
        MultiStream: s2
        phases: ('g', 'l'), T: 298.15 K, P: 101325 Pa
        flow (kg/hr): (l) Water  20
        
        Reset and copy all flows except water:
        
        >>> s2.empty()
        >>> s2.copy_flow(s1, IDs='Water', exclude=True)
        >>> s2.show(flow='kg/hr')
        MultiStream: s2
        phases: ('g', 'l'), T: 298.15 K, P: 101325 Pa
        flow (kg/hr): (l) Ethanol  10
        
        Cut and paste flows:
        
        >>> s2.copy_flow(s1, remove=True)
        >>> s2.show(flow='kg/hr')
        MultiStream: s2
        phases: ('g', 'l'), T: 298.15 K, P: 101325 Pa
        flow (kg/hr): (l) Water    20
                          Ethanol  10
        >>> s1.show()
        MultiStream: s1
        phases: ('g', 'l'), T: 298.15 K, P: 101325 Pa
         flow: 0
         
        The other stream can also be a single phase stream (doesn't have to be a MultiStream object):
                                                            
        Initialize streams:
        
        >>> import thermosteam as tmo
        >>> tmo.settings.set_thermo(['Water', 'Ethanol'], cache=True) 
        >>> s1 = tmo.Stream('s1', Water=20, Ethanol=10, units='kg/hr')
        >>> s2 = tmo.MultiStream('s2')
        
        Copy all flows:
        
        >>> s2.copy_flow(s1)
        >>> s2.show(flow='kg/hr')
        MultiStream: s2
        phases: ('g', 'l'), T: 298.15 K, P: 101325 Pa
        flow (kg/hr): (l) Water    20
                          Ethanol  10
        
        Reset and copy just water flow:
        
        >>> s2.empty()
        >>> s2.copy_flow(s1, IDs='Water')
        >>> s2.show(flow='kg/hr')
        MultiStream: s2
        phases: ('g', 'l'), T: 298.15 K, P: 101325 Pa
        flow (kg/hr): (l) Water  20
        
        Reset and copy all flows except water:
        
        >>> s2.empty()
        >>> s2.copy_flow(s1, IDs='Water', exclude=True)
        >>> s2.show(flow='kg/hr')
        MultiStream: s2
        phases: ('g', 'l'), T: 298.15 K, P: 101325 Pa
        flow (kg/hr): (l) Ethanol  10
        
        Cut and paste flows:
        
        >>> s2.copy_flow(s1, remove=True)
        >>> s2.show(flow='kg/hr')
        MultiStream: s2
        phases: ('g', 'l'), T: 298.15 K, P: 101325 Pa
        flow (kg/hr): (l) Water    20
                          Ethanol  10
        >>> s1.show()
        Stream: s1
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow: 0
        
        """
        if self.chemicals is not other.chemicals and self.chemicals.IDs != other.chemicals.IDs:
            raise ValueError('other stream must have the same chemicals defined to copy flow')
        IDs_index = self.chemicals.get_index(IDs)
        phase_index = self.imol.get_phase_index(phase)
        data = self.imol.data
        other_data = other.imol.data
        multiphase = other_data.ndim == 2
        if exclude:
            data = self.imol.data
            other_data = other.imol.data
            original_data = data.copy()
            if multiphase:
                data[:] = other_data
                data[phase_index, IDs_index] = original_data[phase_index, IDs_index]
                if remove:
                    excluded_data = other_data[phase_index, IDs_index]
                    other_data[:] = 0.
                    other_data[phase_index, IDs_index] = excluded_data
            else:
                other_phase_index = self.imol.get_phase_index(other.phase)
                data[other_phase_index, :] = other_data
                data[phase_index, IDs_index] = original_data[phase_index, IDs_index]
                if remove and (phase is ... or phase_index == other_phase_index):
                    excluded_data = other_data[IDs_index]
                    other_data[:] = 0.
                    other_data[IDs_index] = excluded_data   
        elif multiphase:
            data[phase_index, IDs_index] = other_data[phase_index, IDs_index]
            if remove: other_data[phase_index, IDs_index] = 0.
        else:
            data[:] = 0.
            other_phase_index = self.imol.get_phase_index(other.phase)
            if phase is ... or phase_index == other_phase_index:
                data[other_phase_index, IDs_index] = other_data[IDs_index]
                if remove: other_data[IDs_index] = 0.
    
    def get_normalized_mol(self, IDs: Sequence[str]):
        """
        Return normalized molar fractions of given chemicals. The sum of the result is always 1.

        Parameters
        ----------
        IDs : 
            IDs of chemicals to be normalized.

        Examples
        --------
        >>> import thermosteam as tmo
        >>> tmo.settings.set_thermo(['Water', 'Ethanol', 'Methanol'], cache=True) 
        >>> s1 = tmo.MultiStream('s1', l=[('Water', 20), ('Ethanol', 10), ('Methanol', 10)], units='kmol/hr')
        >>> s1.get_normalized_mol(('Water', 'Ethanol'))
        array([0.667, 0.333])

        """
        z = self.imol[..., IDs].sum(0)
        z /= z.sum()
        return z
    
    def get_normalized_vol(self, IDs: Sequence[str]):
        """
        Return normalized mass fractions of given chemicals. The sum of the result is always 1.

        Parameters
        ----------
        IDs : 
            IDs of chemicals to be normalized.

        Examples
        --------
        >>> import thermosteam as tmo
        >>> tmo.settings.set_thermo(['Water', 'Ethanol', 'Methanol'], cache=True)
        >>> s1 = tmo.MultiStream('s1', l=[('Water', 20), ('Ethanol', 10), ('Methanol', 10)], units='m3/hr')
        >>> s1.get_normalized_vol(('Water', 'Ethanol'))
        array([0.667, 0.333])

        """
        z = self.ivol[..., IDs].sum(0)
        z /= z.sum()
        return z
    
    def get_normalized_mass(self, IDs: Sequence[str]):
        """
        Return normalized mass fractions of given chemicals. The sum of the result is always 1.

        Parameters
        ----------
        IDs : 
            IDs of chemicals to be normalized.

        Examples
        --------
        >>> import thermosteam as tmo
        >>> tmo.settings.set_thermo(['Water', 'Ethanol', 'Methanol'], cache=True) 
        >>> s1 = tmo.MultiStream('s1', l=[('Water', 20), ('Ethanol', 10), ('Methanol', 10)], units='kg/hr')
        >>> s1.get_normalized_mass(('Water', 'Ethanol'))
        array([0.667, 0.333])

        """
        z = self.imass[..., IDs].sum(0)
        z /= z.sum()
        return z
    
    def get_molar_composition(self, IDs: Sequence[str]):
        """
        Return molar composition of given chemicals.

        Parameters
        ----------
        IDs : 
            IDs of chemicals.

        Examples
        --------
        >>> import thermosteam as tmo
        >>> tmo.settings.set_thermo(['Water', 'Ethanol', 'Methanol'], cache=True) 
        >>> s1 = tmo.MultiStream('s1', l=[('Water', 20), ('Ethanol', 10), ('Methanol', 10)], units='kmol/hr')
        >>> s1.get_molar_composition(('Water', 'Ethanol'))
        array([0.5 , 0.25])

        """
        return self.imol[..., IDs].sum(0)/self.F_mol
    
    def get_mass_composition(self, IDs: Sequence[str]):
        """
        Return mass composition of given chemicals.

        Parameters
        ----------
        IDs : 
            IDs of chemicals.

        Examples
        --------
        >>> import thermosteam as tmo
        >>> tmo.settings.set_thermo(['Water', 'Ethanol', 'Methanol'], cache=True) 
        >>> s1 = tmo.MultiStream('s1', l=[('Water', 20), ('Ethanol', 10), ('Methanol', 10)], units='kg/hr')
        >>> s1.get_mass_composition(('Water', 'Ethanol'))
        array([0.5 , 0.25])

        """
        return self.imass[..., IDs].sum(0)/self.F_mass
    
    def get_volumetric_composition(self, IDs: Sequence[str]):
        """
        Return volumetric composition of given chemicals.

        Parameters
        ----------
        IDs : 
            IDs of chemicals.

        Examples
        --------
        >>> import thermosteam as tmo
        >>> tmo.settings.set_thermo(['Water', 'Ethanol', 'Methanol'], cache=True) 
        >>> s1 = tmo.MultiStream('s1', l=[('Water', 20), ('Ethanol', 10), ('Methanol', 10)], units='m3/hr')
        >>> s1.get_volumetric_composition(('Water', 'Ethanol'))
        array([0.5 , 0.25])

        """
        return self.ivol[..., IDs].sum(0)/self.F_vol
    
    def get_concentration(self, phase: str, IDs: Sequence[str]|str, units: Optional[str]=None):
        """
        Return concentration of given chemicals in kmol/m3.

        Parameters
        ----------
        phase :
        IDs : 
            IDs of chemicals.

        Examples
        --------
        >>> import thermosteam as tmo
        >>> tmo.settings.set_thermo(['Water', 'Ethanol', 'Methanol'], cache=True) 
        >>> s1 = tmo.MultiStream('s1', l=[('Water', 20), ('Ethanol', 10), ('Methanol', 10)], units='m3/hr')
        >>> s1.get_concentration('l', ('Water', 'Ethanol'))
        array([27.673,  4.261])

        >>> s1.get_concentration('l', ('Water', 'Ethanol'), 'g/L')
        array([498.532, 196.291])

        """
        F_vol = self.F_vol
        if F_vol == 0.: return 0.
        if units is None:
            return self.imol[IDs] / F_vol 
        else:
            num, denum = units.split('/')
            return self.get_flow(num + '/hr', (phase, IDs)) / self.get_total_flow(denum + '/hr')
        
        return self.imol[phase, IDs] / self.F_vol
    
    ### Equilibrium ###
    
    # @property
    # def equilibrium(self):
    #     """[Equilibrium] An object that can perform equilibrium on the stream."""
    #     phases = self.phases
    #     if phases == ('g', 'l'):
    #         cache = self._vle_cache
    #     elif phases == ('l', 'L'):
    #         cache = self._lle_cache
    #     elif phases == ('l', 's'):
    #         cache = self._sle_cache
    #     else:
    #         return None
    #     return cache.retrieve()
        
    @property
    def vle(self) -> eq.VLE:
        """An object that can perform vapor-liquid equilibrium on the stream."""
        phases = self.phases
        if 'l' not in phases or 'g' not in phases: 
            self.phases = [*phases, 'l', 'g']
        return self._vle_cache.retrieve()
    @property
    def lle(self) -> eq.LLE:
        """An object that can perform liquid-liquid equilibrium on the stream."""
        phases = self.phases
        if 'l' not in phases or 'L' not in phases: 
            self.phases = [*phases, 'l', 'L']
        return self._lle_cache.retrieve()
    @property
    def sle(self) -> eq.SLE:
        """An object that can perform solid-liquid equilibrium on the stream."""
        phases = self.phases
        if 's' not in phases or 'l' not in phases: 
            self.phases = [*phases, 's', 'l']
        return self._sle_cache.retrieve()
    
    ### Casting ###
    
    def as_stream(self):
        """Convert MultiStream to Stream."""
        phase = self.phase
        N_phase = len(phase)
        if N_phase == 1:
            self.phase = phase
        elif N_phase == 0:
            self.phase = self.phases[0]
        else:
            raise RuntimeError('multiple phases present; cannot convert to single phase stream')
    
    def reduce_phases(self):
        """Remove empty phases."""
        self.phase = self.phase
    
    @property
    def phase(self) -> str:
        imol = self._imol
        return ''.join([phases[0] for phases in ('g', 'lL', 'sS') if not imol.phases_are_empty(phases)])
    @phase.setter
    def phase(self, phase):
        N_phase = len(phase)
        if N_phase > 1: 
            self.phases = phase
        else:
            if N_phase == 0: phase = 'l' # Default phase for streams
            self._imol = self._imol.to_chemical_indexer(phase)
            self._streams.clear()
            self.__class__ = Stream
    
    ### Representation ###
    
    def _info_str(self, units, notation, composition, N_max, all_IDs, indexer, factor):
        """Return string with all specifications."""
        basic_info = self._basic_info()
        basic_info += Stream._info_phaseTP(self, self.phases, units, notation)
        flow_units = units['flow']
        flow_notation = notation['flow']
        N_all_IDs = len(all_IDs)
        if N_all_IDs == 0:
            return basic_info + ' flow: 0' 

        # Length of chemical column
        all_lengths = [len(i) for i in all_IDs[:N_max]]
        maxlen = max(all_lengths) + 2
        
        if composition:
            first_line = "composition (%):"
        else:
            first_line = f'flow ({flow_units}):'
        first_line_spaces = len(first_line)*" "

        # Set up chemical data for all phases
        phases_flow_rates_info = ''
        for phase in self.phases:
            phase_data = factor * indexer[phase, all_IDs] 
            IDs, data = nonzeros(all_IDs, phase_data)
            data = np.array(data)
            if not IDs: continue
            if composition:
                total_flow = data.sum()
                data = 100 * data/total_flow
        
            # Get basic structure for phase data            
            beginning = (first_line or first_line_spaces) + f' ({phase}) '
            first_line = False
            new_line = '\n' + len(beginning) * ' '

            # Set chemical data
            flow_rates = ''
            N_IDs = len(data)
            lengths = [len(i) for i in IDs]
            too_many_chemicals = N_IDs > N_max
            N = N_max if too_many_chemicals else N_IDs
            for i in range(N):
                spaces = ' ' * (maxlen - lengths[i])
                if i: flow_rates += new_line    
                flow_rates += f'{IDs[i]}' + spaces + f'{data[i]:{flow_notation}}'
            if too_many_chemicals:
                spaces = ' ' * (maxlen - 3)
                flow_rates += new_line + '...' + spaces + f'{data[N_max:].sum():{flow_notation}}'
            if composition:
                dashes = '-' * (maxlen - 2)
                flow_rates += f"{new_line}{dashes}  {total_flow:{flow_notation}} {flow_units}"
            # Put it together
            phases_flow_rates_info += beginning + flow_rates + '\n'
            
        return basic_info + phases_flow_rates_info.rstrip('\n')
    
    def print(self):
        """
        Print in a format that you can use recreate the stream.

        Examples
        --------
        >>> import thermosteam as tmo
        >>> tmo.settings.set_thermo(['Water', 'Ethanol'], cache=True)
        >>> s1 = tmo.MultiStream(ID='s1',T=298.15, P=101325,
        ...                      l=[('Water', 20), ('Ethanol', 10)], units='kg/hr')
        >>> s1.print()
        MultiStream(ID='s1', phases=('g', 'l'), T=298.15, P=101325, l=[('Water', 1.11), ('Ethanol', 0.2171)])
        
        """
        
        IDs = self.chemicals.IDs
        phase_data = []
        for phase, data in self.imol:
            IDdata = utils.repr_couples(", ", IDs, data)
            if IDdata:
                phase_data.append(f"{phase}=[{IDdata}]")
        dlim = ", "
        phase_data = dlim.join(phase_data)
        phases = f'phases={self.phases}'
        if phase_data:
            phase_data = dlim + phase_data
        price = utils.repr_kwarg('price', self.price)
        print(f"{type(self).__name__}(ID={repr(self.ID)}, {phases}, T={self.T:.2f}, "
              f"P={self.P:.6g}{price}{phase_data})")
    



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
This module includes arbitrary classes and functions.

"""
from __future__ import annotations
import thermosteam as tmo
from math import ceil
import numpy as np
from collections.abc import Iterable
from math import floor, log10
from inspect import signature
from types import FunctionType

__all__ = (
    'extended_signature',
    'factor', 'checkbounds', 'strtuple',
    'format_title', 'format_unit_name',
    'remove_undefined_chemicals',
    'default_chemical_dict', 'subgroup',
    'repr_subgroups', 'repr_items',
    'list_available_names',
    'dictionaries2array',
    'flattened',
    'fill_like', 
    'getfields', 
    'setfields', 
    'copy_maybe', 
    'get_instance',
    'roundsigfigs',
    'array_roundsigfigs',
    'docround',
)

# %% Function signature

def extended_signature(f, g):
    if hasattr(f, '__wrapped__'): f = f.__wrapped__
    sigf = signature(f)
    paramsf = [*sigf.parameters.values()][1:-1]
    sigg = signature(g)
    paramsg = [*sigg.parameters.values()][1:]
    h = FunctionType(f.__code__, f.__globals__, name=f.__name__,
                     argdefs=f.__defaults__, closure=f.__closure__)
    h.__kwdefaults__ = f.__kwdefaults__
    all_params = [*paramsf, *[i.replace(kind=3) for i in paramsg]]
    params = []
    names = set()
    for i in tuple(all_params):
        if i.name in names: continue
        names.add(i.name)
        params.append(i)
    h.__signature__ = sigf.replace(parameters=params)
    h.__annotations__ = f.__annotations__ | g.__annotations__
    h.__wrapped__ = f
    return h

# %% Helpful tools

def flattened(lst):
    flat = []
    isa = isinstance
    for i in lst:
        if isa(i, Iterable):
            flat.extend(flattened(i))
        else:
            flat.append(i)
    return flat

def fill_like(A, B, fields):
    setfield = setattr
    getfield = getattr
    for i in fields: setfield(A, i, getfield(B, i))
    
def getfields(obj, fields, getfield=getattr):
    return [getfield(obj, i) for i in fields]

def setfields(obj, names, fields, setfield=setattr):
    for i,j in zip(names, fields): setfield(obj, i, j)

def copy_maybe(obj):
    return obj.copy() if hasattr(obj, 'copy') else obj

def get_instance(iterable, cls):
    """Return object that is an instance of given class."""
    isa = isinstance
    objs = [i for i in iterable if isa(i, cls)]
    N = len(objs)
    if N == 1:
        return objs[0]
    elif N == 0: # pragma: no cover
        raise ValueError('instance not found')
    else: # pragma: no cover
        raise ValueError('multiple instances found')
            

# %% Number functions

def factor(base_units, new_units):
    if base_units == new_units: return 1
    else: return tmo._Q(1, base_units).to(new_units).magnitude

def checkbounds(x, bounds):
    return bounds[0] < x < bounds[1]
    
def roundsigfigs(x, sigfigs=2, index=1):
    if isinstance(x, Iterable):
        if isinstance(x, str): return x
        n = sigfigs - int(floor(log10(abs(x[index])))) - 1 if abs(x[index]) > 1e-12 else 0.
        try:
            value = np.round(x, n)
        except:
            return np.array(x, dtype=int)
        if (np.array(value, int) == value).all():
            return np.array(value, int)
        else:
            return value
    else:
        n = sigfigs - int(floor(log10(abs(x)))) - 1 if abs(x) > 1e-12 else 0.
        try:
            value = round(x, n)
        except:
            return int(x)
        if int(value) == value:
            return int(value)
        else:
            return value

def docround(obj, n=4):
    """Function to round recursively."""
    if isinstance(obj, np.ndarray):
        return np.round(obj, n)
    elif isinstance(obj, Iterable):
        values = [docround(i, n) for i in obj]
        return obj.__class__(values)
    else:
        return round(obj, n)   

def array_roundsigfigs(arr, sigfigs=2, index=1, inplace=False):
    if not inplace: arr = arr.copy()
    for idx, x in np.ndenumerate(arr):
        if isinstance(x, str): 
            try:
                x = float(x)
            except:
                arr[idx] = x
                continue
        n = sigfigs - int(floor(log10(abs(x)))) - 1 if abs(x) > 1e-12 else 0.
        try:
            value = round(x, n)
        except:
            arr[idx] = int(x)
            continue
        if int(value) == value:
            arr[idx] = int(value)
        else:
            arr[idx] = value
    return arr

# %% Array formulation

def dictionaries2array(dictionaries):
    keys = []
    shape = ()
    for dct in dictionaries: 
        keys.extend(dct)
        for sample in dct.values():
            sample = np.asarray(sample)
            shape = sample.shape
            break
    keys = (*frozenset(keys),)
    index = {j: i for i, j in enumerate(keys)}
    array = np.zeros([*shape, len(dictionaries), len(keys)], dtype=float)
    for i, dct in enumerate(dictionaries):
        for key, value in dct.items():
            j = index[key]
            array[..., i, j] = value
    return array, keys


# %% String functions

def list_available_names(names):
    *names, last = list(names)
    names = ', '.join([repr(i) for i in names]) + ', and ' + repr(last)
    return names

def strtuple(iterable):
    """Return string of all items in the tuple""" 
    string = ''
    function = type(strtuple)
    for i in iterable:
        if isinstance(i , function):
            string += i.__name__ + ', '
        else:
            string += str(i) + ', '
    string = string.rstrip(', ')
    string = '(' + string + ')'
    return string
        
def format_title(line):
    line = line.replace('_', ' ')
    words = []
    word = ''
    last = ''
    for i in line:
        if i.isupper() and last.isalpha() and not last.isupper():
            words.append(word)
            word = i
        else:
            word += i
        last = i
    words.append(word)
    line = ''
    for word in words:
        N_letters = len(word)
        if N_letters > 1:
            line += word + ' '
        else:
            line += word
    line = line.strip(' ')
    first_word, *rest = line.split(' ')
    words = [first_word[0].capitalize() + first_word[1:]]
    for word in rest:
        if not all([(i.isupper() or not last.isalpha()) for i in word]):
            word = word.lower()
        words.append(word)
    return ' '.join(words)

def format_unit_name(name):
    return ''.join([i[0].capitalize() + i[1:] for i in name.split(' ')])
    
def subgroup(items, size=5):
    return [items[size*i: size*(i+1)] for i in range(int(ceil(len(items) / size)))]

def repr_subgroups(subgroups):
    return [', '.join([str(i) for i in j]) for j in subgroups]

def repr_items(start, items, subgroup_size=5, brackets=None):
    N_spaces = len(start)
    if brackets:
        left, right = brackets
        N_spaces += 1
    else:
        left = ''; right = ''
    subgroups = repr_subgroups(subgroup(items, subgroup_size))
    dlim = ",\n" + " " * N_spaces
    return start + left + dlim.join(subgroups) + right

# %% Chemical management

def remove_undefined_chemicals(data: dict, chemicals: tmo.Chemicals):
    for i in tuple(data):
        if i not in chemicals: del data[i]

def default_chemical_dict(dct, chemicals, g, l, s, n=None):
    if n is None: n = l
    for i in chemicals:
        ID = i.ID
        if ID not in dct:
            locked_state = i.locked_state
            if locked_state == 'g': dct[ID] = g
            elif locked_state == 'l': dct[ID] = l
            elif locked_state == 's': dct[ID] = s
            elif locked_state is None: dct[ID] = n
            else: raise RuntimeError(f"unknown locked state '{locked_state}' of chemical '{i}'")



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from chemicals import *
from chemicals.utils import *
from fluids.core import Pr, alpha
from thermosteam.base import functor
from numba import njit
import numpy as np

@functor
def horner(T, coeffs):
    tot = 0
    for c in coeffs: tot = tot * T + c
    return tot 

@njit(cache=True)
def normalize(array, sum_array=None, minimum=1e-16):
    """
    Return a normalized array to a magnitude of 1.
    If magnitude is zero, all fractions will have equal value.
    """
    if sum_array is None: sum_array = array.sum()
    if sum_array < minimum:
        size = array.size
        return np.ones(size)/size
    else:
        return array/sum_array
    
def first_true_index(x):
    """
    Return index of first true value.
    
    """
    for index, i in enumerate(x):
        if i: return index
    
@njit(cache=True)
def mixing_simple(z, y):
    r'''
    Return a weighted average of `y` given the weights, `z`.
    
    Examples
    --------
    >>> import numpy as np
    >>> mixing_simple(np.array([0.1, 0.9]), np.array([0.01, 0.02]))
    0.019000000000000003
    
    '''
    return (z * y).sum()

@njit(cache=True)
def mixing_logarithmic(z, y):
    r'''
    Return the logarithmic weighted average `y` given weights, `z`.
    
    .. math::
    
        y = \sum_i z_i \cdot \log(y_i)
    
    Notes
    -----
    Does not work on negative values.
    Examples
    --------
    >>> import numpy as np
    >>> mixing_logarithmic(np.array([0.1, 0.9]), np.array([0.01, 0.02]))
    0.01866065983073615
    
    '''
    return np.exp((z*np.log(y)).sum())

@njit(cache=True)
def mu_to_nu(mu, rho):
    r"""
    Return the kinematic viscosity (nu) given the dynamic viscosity (mu) and 
    density (rho).
    
    .. math::
    
        \nu = \frac{\mu}{\rho}
    
    Examples
    --------
    >>> mu_to_nu(0.000998, 998.)
    1.0e-06
    
    """
    return mu/rho

@njit(cache=True)
def V_to_rho(V, MW):
    r'''
    Return the density (rho) in kg/m^3 given the molar volume (V) in
    m^3/mol and molecular weight (MW) in g/mol.
    
    .. math::
        \rho = \frac{MW}{1000\cdot V}
    
    Parameters
    ----------
    V : float
        Molar volume, [m^3/mol]
    MW : float
        Molecular weight, [g/mol]
    
    Returns
    -------
    rho : float
        Density, [kg/m^3]
    
    Examples
    --------
    >>> V_to_rho(0.000132, 86.18)
    652.878...
    
    '''
    return MW/V/1000.

@njit(cache=True)
def rho_to_V(rho, MW):
    r'''
    Return the molar volume (V) in m^3/mol given the density (rho) in
    kg/m^3 and molecular weight (MW) in g/mol.
    
    .. math::
        V = \left(\frac{1000 \rho}{MW}\right)^{-1}
    
    Parameters
    ----------
    rho : float
        Density, [kg/m^3]
    MW : float
        Molecular weight, [g/mol]
    
    Returns
    -------
    V : float
        Molar volume, [m^3/mol]
    
    Examples
    --------
    >>> rho_to_V(652.9, 86.18)
    0.0001319957...
    
    '''
    return MW/rho/1000.

def remove_negligible_negative_values(material):
    negative_index = material.negative_index()
    if negative_index:
        material_sum = abs(material).sum()
        if material_sum > 1e-16:
            negligible = material[negative_index] / material_sum > -1e-16
            material[negligible] = 0. 
        else:
            material[negative_index] = 0. 

del njit